import datetime
import bcrypt
import jwt
import uuid
from .models import User
from .. import config
from ..constants import responses


def password_hash(password: str) -> str:
    """
    Args:
        password (str)

    Returns:
        hashed (str)
    """
    password = password.encode()
    hashed = bcrypt.hashpw(password, bcrypt.gensalt())
    return hashed.decode()


def password_match(password: str, hashed: str) -> bool:
    """
    Args:
        password (str)
        hashed (str)

    Returns:
        (bool)
    """
    password = password.encode()
    hashed = hashed.encode()
    return bcrypt.checkpw(password, hashed)


class Payload:
    """
    Attributes:
        error (str)
        user_id (str)
    """
    def __init__(
        self,
        error: str = None,
        user_id: str = None,
        _uuid: str = None
    ):
        self.error = error
        self.user_id = user_id
        self.uuid = _uuid


class AccessTokens:
    """
    Args:
        token_type (str)
        access_token (str)
        refresh_token (str)
        expires_in (int): Tiempo en minutos para que expire el token
    """
    def __init__(
        self,
        token_type: str = None,
        access_token: str = None,
        refresh_token: str = None,
        expires_in: int = None
    ):
        self.token_type = token_type
        self.access_token = access_token
        self.refresh_token = refresh_token
        self.expires_in = expires_in


def token_generate(user: User) -> (AccessTokens, dict):
    """Genera un juego de access token y refresh token

    Args:
        user (User)

    Returns:
        (AccessTokens, dict):
    """
    expires_in = config.TOKEN_EXPIRE_MINUTES

    def token(payload):
        return jwt.encode(
            payload,
            config.SECRET_KEY,
            algorithm='HS256'
        ).decode()

    now = datetime.datetime.utcnow()
    expire = now + datetime.timedelta(minutes=expires_in)

    # access_token
    access_payload = {
        'exp': expire,
        'iat': now,
        'sub': str(user.id),
        'uuid': str(uuid.uuid1())
    }
    access_token = token(access_payload)

    # refresh_token
    refresh_payload = {
        'exp': expire,
        'uuid': str(uuid.uuid1())
    }
    refresh_token = token(refresh_payload)

    return AccessTokens(
        token_type=config.TOKEN_TYPE,
        access_token=access_token,
        refresh_token=refresh_token,
        expires_in=expires_in
    ), {
        'uuid_access': access_payload['uuid'],
        'uuid_refresh': refresh_payload['uuid']
    }


def token_decode(token: str) -> Payload:
    """
    Args:
        token (str)

    Returns:
        Payload:
    """
    try:
        payload = jwt.decode(
            token,
            config.SECRET_KEY,
            algorithms=['HS256']
        )

        return Payload(
            user_id=payload['sub'],
            _uuid=payload['uuid']
        )
    except jwt.ExpiredSignatureError:
        return Payload(responses.SIGNATURE_EXPIRED)
    except Exception:
        return Payload(responses.INVALID_TOKEN)


def refresh_token_decode(refresh_token: str) -> Payload:
    """
    Args:
        refresh_token (str)

    Returns:
        Payload:
    """
    try:
        payload = jwt.decode(
            refresh_token,
            config.SECRET_KEY,
            algorithms=['HS256']
        )

        return Payload(
            _uuid=payload['uuid']
        )
    except jwt.ExpiredSignatureError:
        return Payload(responses.SIGNATURE_EXPIRED)
    except Exception:
        return Payload(responses.INVALID_TOKEN)



from unittest import TestCase
from flask_auth_service_mongo.utils.contract import LoggingInterface


class TestContract(TestCase):

    def test_ok(self):
        log_interface = LoggingInterface()
        msg = "Any message"
        log_interface.debug(msg)
        log_interface.info(msg)
        log_interface.error(msg)



from unittest import TestCase
from flask_auth_service_mongo.security.commands import (
    role_new,
    user_new,
    clear_tokens
)


class TestRoleNew(TestCase):

    def test_params(self):
        context = role_new.make_context(
            'role-new', [
                '-n', 'a name',
                '-p', '{"key": "value"}'
            ]
        )

        self.assertEqual(context.params['name'], 'a name')
        self.assertEqual(context.params['permissions'], '{"key": "value"}')


class TestUserNew(TestCase):

    def test_params(self):
        context = user_new.make_context(
            'new', [
                '-r', 'a role',
                '-u', 'an username',
                '--password', 'a pass'
            ]
        )

        self.assertEqual(context.params['role'], 'a role')
        self.assertEqual(context.params['username'], 'an username')
        self.assertEqual(context.params['password'], 'a pass')


class TestClearTokens(TestCase):

    def test_params(self):
        context = clear_tokens.make_context('clear-tokens', ['-f'])
        self.assertEqual(context.params['forced'], True)

        context = clear_tokens.make_context('clear-tokens', [])
        self.assertEqual(context.params['forced'], False)



from unittest import TestCase
from flask_auth_service_mongo.constants.security import Roles


class TestRoles(TestCase):

    def test_values(self):
        roles = Roles.values()

        self.assertEqual(roles, [
            'is_authenticated',
            'is_anonymously'
        ])



from .user import Response
from .user import SaveUserResponse
from .user import CreateUser
from .user import EditUser
from .user import EditPasswordUser
from .user import DeleteUser
from .user import ResetPassword
from .user import ResetPasswordResponse
from .role import SaveRoleResponse
from .role import NewRole
from .session import Login
from .session import Logout
from .session import RefreshToken
from .session import AuthResponse

__all__ = (
    'Response',
    'SaveUserResponse',
    'CreateUser',
    'EditUser',
    'EditPasswordUser',
    'DeleteUser',
    'SaveRoleResponse',
    'NewRole',
    'ResetPassword',
    'ResetPasswordResponse',
    'Login',
    'Logout',
    'RefreshToken',
    'AuthResponse',
)



from unittest import TestCase
from unittest.mock import patch, Mock
from flask_auth_service_mongo import api_rest
from flask_auth_service_mongo.security.use_cases import AuthResponse
from flask_auth_service_mongo.security.utils import AccessTokens


def resolve_patch(path: str) -> str:
    return f'flask_auth_service_mongo.security.api_rest.session.{path}'


class TestLogin(TestCase):

    @patch(resolve_patch('request'))
    @patch(resolve_patch('Login'))
    def test_error(self, mock_use_case, mock_request):
        # Mocks
        mock_request.get_json.return_value = {'data': 'ok'}
        mock_handle = Mock()
        mock_handle.handle.return_value = AuthResponse(
            message='error',
            http_code=400
        )
        mock_use_case.return_value = mock_handle

        # Exec
        result = api_rest.login('test')

        # Asserts
        self.assertEqual(result, ({'message': 'error'}, 400))
        mock_handle.handle.assert_called_with(
            role='test',
            request={'data': 'ok'}
        )

    @patch(resolve_patch('request'))
    @patch(resolve_patch('Login'))
    def test_ok(self, mock_use_case, mock_request):
        # Mocks
        mock_request.get_json.return_value = {'data': 'ok'}
        mock_handle = Mock()
        mock_handle.handle.return_value = AuthResponse(
            message='ok',
            http_code=200,
            change_password=False,
            tokens=AccessTokens(
                access_token='Abc123',
                token_type='Bearer',
                refresh_token='Abc',
                expires_in=60
            )
        )
        mock_use_case.return_value = mock_handle

        # Exec
        result = api_rest.login('test')

        # Asserts
        self.assertEqual(
            result,
            ({
                'data': {
                    'access_token': 'Abc123',
                    'change_password': False,
                    'expires_in': 60,
                    'refresh_token': 'Abc',
                    'token_type': 'Bearer'
                },
                'message': 'ok'
            }, 200)
        )
        mock_handle.handle.assert_called_with(
            role='test',
            request={'data': 'ok'}
        )


class TestLogout(TestCase):

    @patch(resolve_patch('request'),
           headers={'Authorization': 'Bearer token_test'})
    @patch(resolve_patch('Logout'))
    def test_ok(self, mock_use_case, mock_request):
        # Mocks
        mock_handle = Mock()
        mock_handle.handle.return_value = AuthResponse(
            message='ok',
            http_code=200
        )
        mock_use_case.return_value = mock_handle

        # Exec
        result = api_rest.logout()

        # Asserts
        self.assertEqual(result, ({'message': 'ok'}, 200))
        mock_handle.handle.assert_called_with(
            token='token_test'
        )



from unittest import TestCase
from unittest.mock import patch, Mock
from flask_auth_service_mongo.security.models import User, Role
from flask_auth_service_mongo.security.middlewares import MutationMiddleware


def resolve_patch(path: str) -> str:
    return f'flask_auth_service_mongo.security.middlewares.{path}'


class TestMutationMiddleware(TestCase):

    def middleware(self):
        return MutationMiddleware([
            {
                'mutation': 'login',
                'roles': ['is_anonymously']
            },
            {
                'mutation': 'logout',
                'roles': ['is_authenticated']
            },
            {
                'mutation': 'create_obj',
                'roles': ['custom_role']
            }
        ])

    def test_not_mutation(self):
        # Mocks
        mock_next = Mock(return_value='not_mutation')
        mock_info = Mock(
            operation=Mock(
                operation='other'
            )
        )

        # Exec
        middleware = self.middleware()
        result = middleware.resolve(mock_next, None, mock_info)

        # Asserts
        self.assertEqual(result, 'not_mutation')
        mock_next.assert_called_with(None, mock_info)

    def test_root_is_defined(self):
        # Mocks
        mock_next = Mock(return_value='root_is_defined')
        mock_info = Mock(
            operation=Mock(
                operation='mutation'
            )
        )

        # Exec
        middleware = self.middleware()
        result = middleware.resolve(mock_next, 'root', mock_info)

        # Asserts
        self.assertEqual(result, 'root_is_defined')
        mock_next.assert_called_with('root', mock_info)

    def test_access_control_nod_found(self):
        # Mocks
        mock_next = Mock(return_value='nod_found')
        mock_info = Mock(
            field_name='delete_obj',
            operation=Mock(
                operation='mutation'
            )
        )

        # Exec
        middleware = self.middleware()
        result = middleware.resolve(mock_next, None, mock_info)

        # Asserts
        self.assertEqual(result, 'nod_found')
        mock_next.assert_called_with(None, mock_info)

    @patch(resolve_patch('current_user'))
    def test_user_anonymously(self, mock_current_user):
        # Mocks
        mock_next = Mock(return_value='user_anonymously')
        mock_info = Mock(
            field_name='login',
            operation=Mock(
                operation='mutation'
            )
        )
        mock_current_user.return_value = None

        # Exec
        middleware = self.middleware()
        result = middleware.resolve(mock_next, None, mock_info)

        # Asserts
        self.assertEqual(result, 'user_anonymously')
        mock_next.assert_called_with(None, mock_info)

    @patch(resolve_patch('current_user'))
    def test_unauthorized(self, mock_current_user):
        # Mocks
        mock_next = Mock(return_value='unauthorized')
        mock_info = Mock(
            field_name='logout',
            operation=Mock(
                operation='mutation'
            )
        )
        mock_current_user.return_value = None

        # Exec
        middleware = self.middleware()
        result = middleware.resolve(mock_next, None, mock_info)

        # Asserts
        self.assertEqual(result, None)
        mock_next.assert_not_called()

    @patch(resolve_patch('current_user'))
    def test_is_authenticated(self, mock_current_user):
        # Mocks
        mock_next = Mock(return_value='is_authenticated')
        mock_info = Mock(
            field_name='logout',
            operation=Mock(
                operation='mutation'
            )
        )
        mock_current_user.return_value = User()

        # Exec
        middleware = self.middleware()
        result = middleware.resolve(mock_next, None, mock_info)

        # Asserts
        self.assertEqual(result, 'is_authenticated')
        mock_next.assert_called_with(None, mock_info)

    @patch(resolve_patch('current_user'))
    def test_role_not_valid(self, mock_current_user):
        # Mocks
        mock_next = Mock(return_value='role_not_valid')
        mock_info = Mock(
            field_name='create_obj',
            operation=Mock(
                operation='mutation'
            )
        )
        mock_current_user.return_value = User(
            role=Role(
                name='other_role'
            )
        )

        # Exec
        middleware = self.middleware()
        result = middleware.resolve(mock_next, None, mock_info)

        # Asserts
        self.assertEqual(result, None)
        mock_next.assert_not_called()

    @patch(resolve_patch('current_user'))
    def test_role_ok(self, mock_current_user):
        # Mocks
        mock_next = Mock(return_value='role_ok')
        mock_info = Mock(
            field_name='create_obj',
            operation=Mock(
                operation='mutation'
            )
        )
        mock_current_user.return_value = User(
            role=Role(
                name='custom_role'
            )
        )

        # Exec
        middleware = self.middleware()
        result = middleware.resolve(mock_next, None, mock_info)

        # Asserts
        self.assertEqual(result, 'role_ok')
        mock_next.assert_called_with(None, mock_info)



import graphene
from graphene.relay import Node
from graphene_mongo import MongoengineObjectType
from graphql import GraphQLError
from ...utils.schema import input_to_dictionary
from .. import auth
from ..models import User as UserModel
from ..use_cases import (
    CreateUser as CreateUserUseCase,
    EditUser as EditUserUseCase,
    DeleteUser as DeleteUserUseCase,
    EditPasswordUser,
    ResetPassword as ResetPasswordUseCase
)


__all__ = (
    'UserAttribute',
    'User',
    'CreateUser',
    'UpdateUser',
    'DeleteUser',
    'UpdatePassword',
    'ResetPassword',
)


class UserAttribute:
    username = graphene.String()
    change_password = graphene.Boolean()


class User(MongoengineObjectType, UserAttribute):
    """User Node"""

    class Meta:
        model = UserModel
        interfaces = (Node,)
        exclude_fields = ('password')


class CreateUserInput(graphene.InputObjectType, UserAttribute):
    """Arguments to create a User."""
    username = graphene.String(required=True)
    password = graphene.String(required=True)
    password_confirmed = graphene.String(required=True)
    role = graphene.String(required=True)


class CreateUser(graphene.Mutation):
    user = graphene.Field(lambda: User)

    class Arguments(UserAttribute):
        input = CreateUserInput(required=True)

    def mutate(self, info, input):
        # transforma el request para el caso de uso
        data = input_to_dictionary(input)
        use_case = CreateUserUseCase()
        result = use_case.handle(data)

        if result.http_code > 299:
            raise GraphQLError(result.message)
        return CreateUser(
            user=result.user
        )


class UpdateUserInput(graphene.InputObjectType):
    """Arguments to update a User."""
    id = graphene.ID(required=True)
    change_password = graphene.Boolean(required=True)


class UpdateUser(graphene.Mutation):
    user = graphene.Field(lambda: User)

    class Arguments(UserAttribute):
        input = UpdateUserInput(required=True)

    def mutate(self, info, input):
        # Transform request for use case
        data = input_to_dictionary(input)
        use_case = EditUserUseCase()
        result = use_case.handle(data)

        if result.http_code > 299:
            raise GraphQLError("{}. {}".format(
                result.message,
                result.errors if result.errors else ""
            ))
        return UpdateUser(
            user=result.user
        )


class DeleteUserInput(graphene.InputObjectType):
    """Arguments to delete a User."""
    id = graphene.ID(required=True)


class DeleteUser(graphene.Mutation):
    ok = graphene.Boolean()

    class Arguments():
        input = DeleteUserInput(required=True)

    def mutate(self, info, input):
        # transforma el request para el caso de uso
        data = input_to_dictionary(input)
        use_case = DeleteUserUseCase()
        result = use_case.handle(data)

        if result.http_code > 299:
            raise GraphQLError(result.message)
        return DeleteUser(
            ok=True
        )


class UpdatePasswordInput(graphene.InputObjectType):
    """Arguments to update a User."""
    current_password = graphene.String(required=True)
    new_password = graphene.String(required=True)
    password_confirmed = graphene.String(required=True)


class UpdatePassword(graphene.Mutation):
    ok = graphene.Boolean()

    class Arguments():
        input = UpdatePasswordInput(required=True)

    def mutate(self, info, input):
        user = auth.current_user()

        # transforma el request para el caso de uso
        data = input_to_dictionary(input)
        data['id'] = str(user.id)

        use_case = EditPasswordUser()
        result = use_case.handle(data)

        if result.http_code > 299:
            raise GraphQLError(result.message)
        return UpdatePassword(
            ok=True
        )


class ResetPasswordInput(graphene.InputObjectType):
    """Arguments to ResetPassword."""
    id = graphene.ID(required=True)


class ResetPassword(graphene.Mutation):
    password = graphene.String()

    class Arguments():
        input = ResetPasswordInput(required=True)

    def mutate(self, info, input):
        data = input_to_dictionary(input)
        use_case = ResetPasswordUseCase()
        result = use_case.handle(data)

        if result.http_code > 299:
            raise GraphQLError(result.message)
        return ResetPassword(
            password=result.password
        )



""" Default config.py file"""
import os


TOKEN_TYPE = 'Bearer'  # nosec
WHITE_LIST_TOKEN = os.environ.get('WHITE_LIST_TOKEN', True)
SECRET_KEY = os.environ.get('SECRET_KEY', 'not-secret')
USERNAME_MIN_LENGTH = os.environ.get('USERNAME_MIN_LENGTH', 3)
PASSWORD_MIN_LENGTH = os.environ.get('PASSWORD_MIN_LENGTH', 3)
TOKEN_EXPIRE_MINUTES = os.environ.get('TOKEN_EXPIRE_MINUTES', 60)
RESET_PASSWORD_LEN_GENERATOR = os.environ.get(
    'RESET_PASSWORD_LEN_GENERATOR', 8)
REQUIRE_PASSWORD_CHANGE = os.environ.get('REQUIRE_PASSWORD_CHANGE', False)



from .role import Role
from .user import User
from .user import CreateUser
from .user import UpdateUser
from .user import DeleteUser
from .user import UpdatePassword
from .user import ResetPassword
from .session import RefreshToken

__all__ = (
    'Role',
    'User',
    'CreateUser',
    'UpdateUser',
    'DeleteUser',
    'UpdatePassword',
    'ResetPassword',
    'RefreshToken',
)



from unittest import TestCase
from flask import Flask
from flask_auth_service_mongo import config
from flask_auth_service_mongo.security.models import User
from flask_auth_service_mongo.security.utils import (
    Payload,
    password_hash,
    password_match,
    token_decode,
    token_generate,
    refresh_token_decode,
    AccessTokens
)


class TestPayload(TestCase):
    def test_new_full(self):
        params = ('error', 'user_id', 'uuid')
        payload = Payload(*params)
        self.assertEqual(params[0], payload.error)
        self.assertEqual(params[1], payload.user_id)
        self.assertEqual(params[2], payload.uuid)

    def test_new_nullable(self):
        payload = Payload()
        self.assertIsNone(payload.error)
        self.assertIsNone(payload.user_id)
        self.assertIsNone(payload.uuid)


class TestPassword(TestCase):
    """Pruebas del generador del hash y el match para la contraseña"""

    def test_hash_and_match(self):
        param = 'password'
        hashpw = password_hash(param)

        self.assertIsInstance(hashpw, str)
        self.assertNotEqual(hashpw, param)

        matchpw = password_match(param, hashpw)

        self.assertIsInstance(matchpw, bool)
        self.assertTrue(matchpw)

        matchpw = password_match('param', hashpw)

        self.assertIsInstance(matchpw, bool)
        self.assertFalse(matchpw)


class TestToken(TestCase):
    """Pruebas para el generador y el decode del token"""

    def test_generate_and_decode(self):
        # Crea el contexto del app flask
        app = Flask(__name__)
        app.config['SECRET_KEY'] = 'not-secret'

        with app.app_context():
            user = User(id='id')

            # generar tokens
            tokens, uuids = token_generate(user)
            self.assertIsInstance(tokens, AccessTokens)
            self.assertEqual(tokens.token_type, 'Bearer')
            self.assertEqual(tokens.expires_in, config.TOKEN_EXPIRE_MINUTES)
            self.assertIsInstance(uuids['uuid_access'], str)
            self.assertIsInstance(uuids['uuid_refresh'], str)
            self.assertNotEqual(uuids['uuid_access'], uuids['uuid_refresh'])

            # Decode Ok access_token
            payload = token_decode(tokens.access_token)
            self.assertIsInstance(payload, Payload)
            self.assertIsNone(payload.error)
            self.assertEqual(payload.user_id, 'id')
            self.assertEqual(payload.uuid, uuids['uuid_access'])

            # Decode Ok refresh_token
            payload = refresh_token_decode(tokens.refresh_token)
            self.assertIsInstance(payload, Payload)
            self.assertIsNone(payload.error)
            self.assertEqual(payload.uuid, uuids['uuid_refresh'])

            # Decode expired token
            token = ('eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJleHAiOjE1NzYwMjE'
                     '5NzYsImlhdCI6MTU3NjAxODM3Niwic3ViIjoiNWRmMDIxYmU1Y2I0Y2Z'
                     'hZDgyMzM5MTg4Iiwicm9sIjoiYWRtaW4ifQ.PYY_7TRhSyd0_H_tXGBt'
                     'WCSm2K_pPuNjyOk4NERgNrk')
            payload = token_decode(token)
            self.assertIsInstance(payload, Payload)
            self.assertEqual(payload.error, 'signature_expired')
            self.assertIsNone(payload.user_id)

            # Decode expired token
            token = 'eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJleHAiOjE1NzYwMjE'
            payload = token_decode(token)
            self.assertIsInstance(payload, Payload)
            self.assertEqual(payload.error, 'invalid_token')
            self.assertIsNone(payload.user_id)



import graphene
from graphql import GraphQLError
from ...utils.schema import input_to_dictionary
from ..use_cases import RefreshToken as RefreshTokenUseCase

__all__ = (
    'RefreshToken',
)


class RefreshTokenInput(graphene.InputObjectType):
    """Arguments to update a User."""
    refresh_token = graphene.String(required=True)


class RefreshToken(graphene.Mutation):
    token_type = graphene.String()
    access_token = graphene.String()
    refresh_token = graphene.String()
    expires_in = graphene.Int()

    class Arguments():
        input = RefreshTokenInput(required=True)

    def mutate(self, info, input):
        data = input_to_dictionary(input)
        use_case = RefreshTokenUseCase()
        result = use_case.handle(data)

        if result.http_code > 299:
            raise GraphQLError("{}. {}".format(
                result.message,
                result.errors if result.errors else ""
            ))

        return RefreshToken(
            token_type=result.tokens.token_type,
            access_token=result.tokens.access_token,
            refresh_token=result.tokens.refresh_token,
            expires_in=result.tokens.expires_in
        )



import abc
from ..constants.enums import HttpCode
from .contract import LoggingInterface


class Response:
    """Respuesta Genérica para los casos de uso

    Attributes:
        _http_code (constants.enums.HttpCode):
            Garantiza que el codigo sea valido. Default HttpCode.OK
        message (str)
        errors (list)
    """

    def __init__(
        self,
        http_code: int = None,
        message: str = None,
        errors: list = None
    ):
        # Default HttpCode
        self._http_code: HttpCode = HttpCode.OK

        self.http_code = http_code
        self.message = message
        self.errors = errors

    @property
    def http_code(self) -> int:
        """int: Código http"""
        if self._http_code:
            return self._http_code.value
        return None

    @http_code.setter
    def http_code(self, http_code) -> None:
        if isinstance(http_code, HttpCode):
            self._http_code = http_code
        elif http_code:
            self._http_code = HttpCode(http_code)

    def __json__(self):
        return {
            'http_code': self.http_code,
            'message': self.message,
            'errors': self.errors,
        }


class UseCaseInterface(metaclass=abc.ABCMeta):
    """Contrato para los casos de uso
    Attributes:
        _path (str)
        _logger (utils.contract.LoggingInterface)
    """
    _path = None
    _logger: LoggingInterface = None

    def _debugger(self, message):
        """Se ejecuta solo si esta activo el debug

        Args:
            message (str)
        """
        self._logger.debug(
            '{}: {}'.format(self._path, message)
            )

    @abc.abstractmethod
    def handle(self, *args, **kwargs) -> Response:
        """Único método público, este debe recibir el request
        y retornar el resultado de todo el caso de uso

        Nota:
            Cualquier otro parámetro que el caso de uso necesite,
            que no llegue en el request de la petición

        Args:
            request (dict): (Opcional) Recomendado
            args: other positional arguments
            kwargs: other keyword arguments

        Returns:
            response (utils.use_case.Response): Instancia de Response o\
            de una clase hija (recomendado si se requiere retornar\
            datos que no sean atributos de Response)
        """
        pass



from cerberus import Validator as ValidatorCerberus
from bson.objectid import ObjectId


class Validator(ValidatorCerberus):

    def _validate_no_spaces(self, no_spaces, field, value):
        """ Test the value must not contain spaces

        The rule's arguments are validated against this schema:
        {'type': 'boolean'}
        """
        if no_spaces and (' ' in value):
            self._error(field, "Must not contain spaces")

    def _validate_mongo_id(self, mongo_id, field, value):
        """ Test the ObjectId of a value.

        The rule's arguments are validated against this schema:
        {'type': 'boolean'}
        """
        if mongo_id and not ObjectId.is_valid(value):
            self._error(field, "Is not a valid ID")

    def _validate_only_numbers(self, only_numbers, field, value):
        """ Test only numbers of a value.

        The rule's arguments are validated against this schema:
        {'type': 'boolean'}
        """

        if only_numbers and not value.isdigit():
            self._error(field, "Only numeric digits is allowed")




class Roles:
    IS_AUTHENTICATED_FULLY = 'is_authenticated'
    IS_AUTHENTICATED_ANONYMOUSLY = 'is_anonymously'

    @staticmethod
    def values() -> list:
        return [
            Roles.IS_AUTHENTICATED_FULLY,
            Roles.IS_AUTHENTICATED_ANONYMOUSLY,
        ]



from ... import config
from ...constants import responses
from ...constants.enums import HttpCode
from ...utils.use_case import Response, UseCaseInterface
from ...utils.validator import Validator
from ..auth import current_user
from ..utils import (
    password_match,
    token_generate,
    token_decode,
    refresh_token_decode,
    AccessTokens
)
from ..models import WhitelistToken
from ..repository import (
    UserRepository,
    WhitelistTokenRepository
)


__all__ = (
    'Login',
    'Logout',
    'RefreshToken',
    'AuthResponse',
)


class AuthResponse(Response):
    """Respuesta para los casos de uso de Auth

    Args:
        tokens (AccessTokens)
        change_password (bool): Indica si es necesario cambiar la contraseña
        http_code (int)
        message (str)
        errors (list)
    """
    def __init__(
        self,
        tokens: str = None,
        change_password: bool = None,
        *args,
        **kwargs
    ):
        self.tokens = tokens
        self.change_password = change_password

        super().__init__(*args, **kwargs)


class BaseAuth(UseCaseInterface):

    def _new_tokens(self, user) -> AccessTokens:
        """Crea un juego de tokens y lo registra en WhitelistToken

        Args:
            user (User)

        Returns:
            AccessTokens:
        """
        tokens, uuids = token_generate(user)

        if config.WHITE_LIST_TOKEN:
            WhitelistToken(
                uuid_access=uuids['uuid_access'],
                uuid_refresh=uuids['uuid_refresh'],
                user=user
            ).save()

        return tokens


class Login(BaseAuth):

    def handle(self, request: dict, role=None) -> AuthResponse:
        """
        Args:
            request (dict) Ejemplo:
                ::
                    {
                        'username': 'username',
                        'password': 'password'
                    }
            role (str|List[str]): or None. if not None: Allowed roles

        Returns:
            AuthResponse:
        """
        if isinstance(role, str):
            self._role = [role]
        else:
            self._role = role
        self._request = request

        valid, errors = self.__request_is_valid()
        if not valid:
            return AuthResponse(
                message=responses.BAD_REQUEST,
                http_code=HttpCode.BAD_REQUEST,
                errors=errors
            )

        username = self._request['username'].lower()
        self._user = UserRepository.find_one(username=username)
        if not self.__credential_conditions():
            return AuthResponse(
                message=responses.BAD_CREDENTIALS,
                http_code=HttpCode.BAD_REQUEST
            )

        return AuthResponse(
            message=responses.OK,
            tokens=self._new_tokens(self._user),
            change_password=self._user.change_password
        )

    def __credential_conditions(self) -> tuple:
        """Condiciones para verificar credenciales.
        Usuario, contraseña y role

        Returns:
            (tuple)
        """
        if self._role:
            return (
                self._user and
                self._user.role.name in self._role and
                password_match(self._request['password'], self._user.password)
            )
        return (
            self._user and
            password_match(self._request['password'], self._user.password)
        )

    def __request_is_valid(self) -> (bool, dict):
        """El request debe cumplir con la estructura,
        no debe tener key adicionales

        Returns:
            (bool, dict): (is_valid, errors)
        """
        v = Validator()
        v.schema = {
            "username": {
                "type": "string",
                "required": True,
                "no_spaces": True
            },
            "password": {
                "type": "string",
                "required": True,
                "no_spaces": True
            }
        }

        valid = v.validate(self._request)
        return valid, v.errors


class RefreshToken(BaseAuth):

    def handle(self, request: dict) -> AuthResponse:
        """
        Args:
            request (dict): Ejemplo:
                ::
                    {
                        'refresh_token': 'a_refresh_token'
                    }

        Returns:
            AuthResponse:
        """
        self.__request = request

        valid, response = self.__validate()
        if not valid:
            return response

        tokens = self._new_tokens(self.__user)
        return AuthResponse(
            message=responses.OK,
            tokens=tokens
        )

    def __validate(self) -> (bool, AuthResponse):
        """Valida el request y carga self.__portfolio y self.__payment

        Returns:
            (bool, AuthResponse): (True, None) Si es valido.
                (False, AuthResponse) Si no es valido.
        """
        valid, errors = self.__request_is_valid()
        if not valid:
            return False, AuthResponse(
                http_code=HttpCode.BAD_REQUEST,
                message=responses.BAD_REQUEST,
                errors=errors
            )

        payload = refresh_token_decode(
            self.__request['refresh_token']
        )
        if payload.error:
            return False, AuthResponse(
                http_code=HttpCode.UNAUTHORIZED,
                message=payload.error
            )

        self.__user = current_user()
        if config.WHITE_LIST_TOKEN:
            self.__whitelist = WhitelistTokenRepository.find_one(
                uuid_refresh=payload.uuid
            )
            if not self.__whitelist:
                return False, AuthResponse(
                    http_code=HttpCode.UNAUTHORIZED,
                    message=responses.UNAUTHORIZED
                )

            if self.__user.id != self.__whitelist.user.id:
                # El usuario autenticado no es propietario del refresh_token
                return False, AuthResponse(
                    http_code=HttpCode.UNAUTHORIZED,
                    message=responses.UNAUTHORIZED
                )

        return True, None

    def __request_is_valid(self) -> (bool, dict):
        """El request debe cumplir con la estructura,
        no debe tener key adicionales

        Returns:
            (bool, dict): (is_valid, errors)
        """
        v = Validator()
        v.schema = {
            "refresh_token": {
                "type": "string",
                "required": True
            }
        }

        valid = v.validate(self.__request)
        return valid, v.errors


class Logout(UseCaseInterface):

    def handle(self, token: str) -> Response:
        """
        Args:
            token (str)

        Returns:
            Response:
        """
        self.__token = token
        self.__delete_token()

        return Response(
            message=responses.OK
        )

    def __delete_token(self) -> None:
        """Borra el token de la WhitelistToken"""
        payload = token_decode(self.__token)
        token = WhitelistTokenRepository.find_one(
            uuid_access=payload.uuid
        )
        if token:
            token.delete()
